//head <-> prev <-> next <-> tail

var Record = function(key, value, expire) {
    this.key = key;
    this.value = value;
    this.expire = expire;
    this.deadline = Date.now() + expire;
    this.next = null;
    this.prev = null;
};

var Cache = function(max, defaultExpire) {
    this.max = max || 10 * 1024 * 1024;
    this.defaultExpire = defaultExpire || 60 * 60 * 1000;
    this.__cache__ = {};
    this.__length__ = 0;
    this.__head__ = null;
    this.__tail__ = null;
};

Cache.prototype.set = function(key, value, expire) {

    var __cache__ = this.__cache__;
    var record = __cache__[key];
    
    if (!record) {
        var expire = expire || this.defaultExpire;
        record = __cache__[key] = {
            key: key,
            value: value,
            expire: expire,
            deadline: Date.now() + expire
        };
        this.unshift(record);

    } else {

        record.expire = expire || record.expire;
        record.deadline = Date.now() + record.expire;
        record.value = value;
        this.remove(record);
        this.unshift(record);

    }

    //this.sort(record);

};

Cache.prototype.get = function(key) {

    var value;
    var record = this.__cache__[key];

    if (record) {
        record.deadline = Date.now() + record.expire;
        value = record.value;
        this.remove(key);
        this.unshift(record);
        //this.sort(record);
    }

    return value;

};

Cache.prototype.remove = function(key) {

    var __cache__ = this.__cache__;
    var record = __cache__[key];
    
    if (!record) return;

    __cache__[key] = null;
    this.__length__ -= 1;

    var recNext = record.next;
    var recPrev = record.prev;

    if (recNext && recPrev) {
        recNext.prev = recPrev;
        recPrev.next = recNext;
        return;
    }

    if (recNext) {
        this.__head__ = recNext;
        recNext.prev = null;
        return;
    }

    if (recPrev) {
        this.__tail__ = recPrev;
        recPrev.next = null;
        return;
    }

    this.head = this.tail = null;

};

Cache.prototype.push = function(record) {

    this.__length__ += 1;
    
    if (!this.__tail__) {
        this.__head__ = this.__tail__ = record;
        return;
    }

    this.__tail__.next = record;
    record.prev = this.__tail__;
    record.next = null;
    this.__tail__ = record;

};

Cache.prototype.pop = function() {

    if (!this.__tail__) return;
    
    var record = this.__tail__;

    if (record.prev) {
        record.prev.next = null;
        this.__tail__ = record.prev;
    } else {
        this.__tail__ = this.__head__ = null;
    }

    this.__length__ -= 1;
    this.__cache__[record.key] = null;
    return record;

};

Cache.prototype.unshift = function(record) {
    this.__length__ += 1;
    if (!this.__head__) {
        this.__head__ = this.__tail__ = record;
        return;
    }
    this.__head__.prev = record;
    record.next = this.__head__;
    record.prev = null;
    this.__head__ = record;    
};

Cache.prototype.shift = function() {

    if (!this.__head__) return;

    var record = this.__head__;

    if (record.next) {
        record.next.prev = null;
        this.__head__ = record.next;
    } else {
        this.__head__ = this.__tail__ = null;
    }

    this.__tail__ -= 1;
    this.__cache__[record.key] = null;
    return record;

};

/*
Cache.prototype.sort = function(record) {
    var direct;

    var recNext = record.next;
    var recPrev = record.prev;
    var deadline = record.deadline;

    if (recNext && recPrev) {
        direct = deadline > recPrev.deadline ? 
            'forward' : 
            (deadline < recNext.deadline ? 'backward' : null);
    } else if (recNext) {
        direct = deadline < recNext.deadline ? 'backward' : null;
    } else if (recPrev) {
        direct = deadline > recPrev.deadline ? 'forward' : null;
    }

    direct && this[direct](record);

};

Cache.prototype.forward = function(record) {
    var recPrev = record.prev;
    var deadline = record.deadline;

    var o = 0;
    while (recPrev && deadline > recPrev.deadline) {
        o++;
        if (!record.next) {
            this.__tail__ = recPrev;
        } else {
            record.next.prev = recPrev;
        }

        if (!recPrev.prev) {
            this.__head__ = record;
        } else {
            recPrev.prev.next = record;
        }

        recPrev.next = record.next;
        record.prev = recPrev.prev;

        record.next = recPrev;
        recPrev.prev = record;

        recPrev = record.prev;
    }
    console.log(o);
};

Cache.prototype.backward = function(record) {
    var recNext = record.next;
    var deadline = record.deadline;

    while (recNext && deadline < recNext.deadline) {

        if (!record.prev) {
            this.__head__ = recNext;
        } else {
            record.prev.next = recNext;
        }

        if (!recNext.next) {
            this.__tail__ = record;
        } else {
            recNext.next.prev = record;
        }

        recNext.prev = record.prev;
        record.next = recNext.next;
        record.prev = recNext;
        recNext.next = record;

        recNext = record.next;
    }
};
*/
Cache.prototype.update = function() {
    var self = this;
    self.__updateTimer__ = setTimeout(function() {
        var now = self.lastUpdate = Date.now();
        while (self.__tail__ && self.__tail__.deadline < now) {
            self.pop();
        }
    }, 60 * 1000);
};

Cache.prototype.print = function() {
    var o = this.__head__;
    while (o) {
        console.log(o.key + ' : ' + o.value + ' | ' + o.expire + ' | ' + o.deadline);
        o = o.next;
    }
};

exports.init = function(max, defaultExpire) {
    return new Cache(max, defaultExpire);
};

